# -*- coding:utf-8 -*-

"""PyHdust auxiliary module: third-part MCMC plotting tools.

:co-author:Dan Foreman-Mackey
:license: GNU GPL v3.0 https://github.com/danmoser/pyhdust/blob/master/LICENSE
"""
from __future__ import print_function, absolute_import, unicode_literals
import numpy as _np
import warnings as _warn
import pyhdust.phc as _phc
import pyhdust.stats as _stt
from six import string_types as _strtypes

try:
    import matplotlib as _mpl
    import matplotlib.pyplot as _plt
    import matplotlib.cm as _cm
    import matplotlib.gridspec as _gridspec 
    from matplotlib.ticker import MaxNLocator as _MaxNLoc
    from scipy.ndimage import gaussian_filter as _gf
    from scipy import stats as _stats
except ImportError:
    gaussian_filter = None
    _warn.warn('matplotlib or scipy not installed!!!')

__author__ = "Daniel Moser"
__email__ = "dmfaes@gmail.com"


def kerncorner(xs, cmapn='gray_r', ncl=3, bestvals='perc', labels=[]):
    """ xs = (niteractions, ndim) 
    ncl = number os contour lines
    bestvals = None, 'peak' or 'perc'
    """
    ndim = _np.shape(xs)[1]
    if len(labels) < ndim:
        labels += ['']*(ndim-len(labels))

    fig = _plt.figure(figsize=(9, 9))
    gs = _gridspec.GridSpec(ndim, ndim)
    gs.update(hspace=0.01)
    fs = dict(_mpl.rcParams.viewitems())['font.size']-4

    axs = []
    xslim = []
    for i in range(ndim):
        ax = []
        xslim += [[_np.min(xs[:, i]), _np.max(xs[:, i])]]
        for j in range(i+1):
            ax += [_plt.subplot(gs[i, j])]
        axs += [ax]

    for i in range(ndim):
        for j in range(i+1):
            axs[i][j].locator_params(nbins=5)
            xmin, xmax = xslim[j]
            if i == j:
                kernel = _stats.gaussian_kde(xs[:, i])
                x = _np.linspace(xmin, xmax, 101)
                y = kernel(x)
                axs[i][j].plot(x, y/_np.max(y))
                axs[i][j].set_xlim([xmin, xmax])
                ymin, ymax = axs[i][j].get_ylim()
                if isinstance(bestvals, _strtypes):
                    if bestvals.startswith('perc'):
                        pc = _np.percentile(xs[:, i], [15.9, 50.0, 84.1])
                    elif bestvals.startswith('peak'): 
                        mode = x[_phc.find_nearest(y, _np.max(y), idx=True)]
                        mad = _stt.mad(xs[:, i])
                        pc = mode + _np.array([-mad, 0, mad])
                    else:
                        _warn.warn('Invalid `bestvals` in kerncorner', 
                            stacklevel=2)
                    if 'pc' in locals():
                        dx = xmax-xmin
                        for p in pc:
                            if p < xmin+0.015*dx:
                                p = xmin+0.015*dx
                            if p > xmax-0.015*dx:
                                p = xmax-0.015*dx
                            axs[i][j].plot([p, p], [ymin, ymax], 'k--')
            else:
                y = xs[:, i]
                x = xs[:, j]
                ymin, ymax = xslim[i]
                #
                values = _np.vstack([x, y])
                kernel = _stats.gaussian_kde(values)
                X, Y = _np.mgrid[xmin:xmax:100j, ymin:ymax:100j]
                positions = _np.vstack([X.ravel(), Y.ravel()])
                Z = _np.reshape(kernel(positions), X.shape)
                #
                # axs[i][j].plot(x, y, 'k.')
                axs[i][j].imshow(Z.T, cmap=_plt.get_cmap(cmapn), 
                    extent=[xmin, xmax, ymin, ymax], 
                    origin='lower',
                )
                if ncl > 0:
                    CS = axs[i][j].contour(X, Y, Z, ncl, 
                    colors=_phc.gradColor(range(ncl), cmapn='copper'))
                # axs[i][j].clabel(CS, inline=1, fontsize=fs)
            if j == 0:
                axs[i][j].set_ylabel(labels[i])
            if i == ndim-1:
                axs[i][j].set_xlabel(labels[j])
            if (i == 0 and j == 0) or (j != 0):
                axs[i][j].set_yticklabels([])
            axs[i][j].set_aspect(abs(xmax-xmin)/abs(ymax-ymin))
            if i != ndim-1:
                # axs[i][j].get_xaxis().set_visible(False)
                # axs[i][j].xaxis.set_visible(False)
                axs[i][j].set_xticklabels([])
            else:
                _plt.setp( axs[i][j].xaxis.get_majorticklabels(), rotation=50 )

    _phc.savefig(fig)  # figname='outname')
    return


def corner(xs, bins=20, range=None, weights=None, color="k",
           smooth=None, smooth1d=None,
           labels=None, label_kwargs=None,
           show_titles=False, title_fmt=".2f", title_kwargs=None,
           truths=None, truth_color="#4682b4",
           scale_hist=False, quantiles=None, verbose=False, fig=None,
           max_n_ticks=5, top_ticks=False, hist_kwargs=None, 
           qtdiff=None, **hist2d_kwargs):
    """
    Make a *sick* corner plot showing the projections of a data set in a
    multi-dimensional space. kwargs are passed to hist2d() or used for
    `matplotlib` styling.

    Parameters
    ----------
    xs : array_like (nsamples, ndim)
        The samples. This should be a 1- or 2-dimensional array. For a 1-D
        array this results in a simple histogram. For a 2-D array, the zeroth
        axis is the list of samples and the next axis are the dimensions of
        the space.

    weights : array_like (nsamples,)
        The weight of each sample. If `None` (default), samples are given
        equal weight.

    labels : iterable (ndim,) (optional)
        A list of names for the dimensions. If a ``xs`` is a
        ``pandas.DataFrame``, labels will default to column names.

    show_titles : bool (optional)
        Displays a title above each 1-D histogram showing the 0.5 quantile
        with the upper and lower errors supplied by the quantiles argument.

    title_fmt : string (optional)
        The format string for the quantiles given in titles.
        (default: `.2f`)

    title_args : dict (optional)
        Any extra keyword arguments to send to the `add_title` command.

    extents : iterable (ndim,) (optional)
        A list where each element is either a length 2 tuple containing
        lower and upper bounds (extents) or a float in range (0., 1.)
        giving the fraction of samples to include in bounds, e.g.,
        [(0.,10.), (1.,5), 0.999, etc.].
        If a fraction, the bounds are chosen to be equal-tailed.

    truths : iterable (ndim,) (optional)
        A list of reference values to indicate on the plots.  Individual
        values can be omitted by using ``None``.

    truth_color : str (optional)
        A ``matplotlib`` style color for the ``truths`` makers.

    scale_hist : bool (optional)
        Should the 1-D histograms be scaled in such a way that the zero line
        is visible?

    quantiles : iterable (optional)
        A list of fractional quantiles to show on the 1-D histograms as
        vertical dashed lines.

    verbose : bool (optional)
        If true, print the values of the computed quantiles.

    plot_contours : bool (optional)
        Draw contours for dense regions of the plot.

    plot_datapoints : bool (optional)
        Draw the individual data points.

    max_n_ticks: int (optional)
        maximum number of ticks to try to use

    fig : matplotlib.Figure (optional)
        Overplot onto the provided figure object.

    """
    if quantiles is None:
        quantiles = []
    if title_kwargs is None:
        title_kwargs = dict()
    if label_kwargs is None:
        label_kwargs = dict()
    if qtdiff is None:
        qtdiff = _np.zeros(xs.shape[1])

    # Try filling in labels from pandas.DataFrame columns.
    if labels is None:
        try:
            labels = xs.columns
        except AttributeError:
            pass

    # Deal with 1D sample lists.
    xs = _np.atleast_1d(xs)
    if len(xs.shape) == 1:
        xs = _np.atleast_2d(xs)
    else:
        assert len(xs.shape) == 2, "The input sample array must be 1- or 2-D."
        xs = xs.T
    assert xs.shape[0] <= xs.shape[1], "I don't believe that you want more " \
                                       "dimensions than samples!"

    # Parse the weight array.
    if weights is not None:
        weights = _np.asarray(weights)
        if weights.ndim != 1:
            raise ValueError("Weights must be 1-D")
        if xs.shape[1] != weights.shape[0]:
            raise ValueError("Lengths of weights must match number of samples")

    # Parse the parameter ranges.
    if range is None:
        if "extents" in hist2d_kwargs:
            _warn.warn("Deprecated keyword argument 'extents'. "
                "Use 'range' instead.")
            range = hist2d_kwargs.pop("extents")
        else:
            range = [[x.min(), x.max()] for x in xs]
            # Check for parameters that never change.
            m = _np.array([e[0] == e[1] for e in range], dtype=bool)
            if _np.any(m):
                raise ValueError(("It looks like the parameter(s) in "
                                  "column(s) {0} have no dynamic range. "
                                  "Please provide a `range` argument.")
                                 .format(", ".join(map(
                                     "{0}".format, _np.arange(len(m))[m]))))

    else:
        # If any of the extents are percentiles, convert them to ranges.
        for i, _ in enumerate(range):
            try:
                emin, emax = range[i]
            except TypeError:
                q = [0.5 - 0.5 * range[i], 0.5 + 0.5 * range[i]]
                range[i] = quantile(xs[i], q, weights=weights)

    if len(range) != xs.shape[0]:
        raise ValueError("Dimension mismatch between samples and range")

    # Parse the bin specifications.
    try:
        bins = [float(bins) for _ in range]
    except TypeError:
        if len(bins) != len(range):
            raise ValueError("Dimension mismatch between bins and range")

    # Some magic numbers for pretty axis layout.
    K = len(xs)
    factor = 2.0           # size of one side of one panel
    lbdim = 0.5 * factor   # size of left/bottom margin
    trdim = 0.2 * factor   # size of top/right margin
    whspace = 0.05         # w/hspace size
    plotdim = factor * K + factor * (K - 1.) * whspace
    dim = lbdim + plotdim + trdim

    # Create a new figure if one wasn't provided.
    if fig is None:
        fig, axes = _plt.subplots(K, K, figsize=(dim, dim), tight_layout=False)
    else:
        try:
            axes = _np.array(fig.axes).reshape((K, K))
        except:
            raise ValueError("Provided figure has {0} axes, but data has "
                             "dimensions K={1}".format(len(fig.axes), K))
        fig.set_tight_layout(False)

    # Format the figure.
    lb = lbdim / dim
    tr = (lbdim + plotdim) / dim
    fig.subplots_adjust(left=lb, bottom=lb, right=tr, top=tr,
                        wspace=whspace, hspace=whspace)

    # Set up the default histogram keywords.
    if hist_kwargs is None:
        hist_kwargs = dict()
    hist_kwargs["color"] = hist_kwargs.get("color", color)
    if smooth1d is None:
        hist_kwargs["histtype"] = hist_kwargs.get("histtype", "step")

    for i, x in enumerate(xs):
        # Deal with masked arrays.
        if hasattr(x, "compressed"):
            x = x.compressed()

        if _np.shape(xs)[0] == 1:
            ax = axes
        else:
            ax = axes[i, i]
        # Plot the histograms.
        if smooth1d is None:
            n, _, _ = ax.hist(x, bins=int(round(bins[i])), weights=weights,
                              range=range[i], **hist_kwargs)
        else:
            if gaussian_filter is None:
                raise ImportError("Please install scipy for smoothing")
            n, b = _np.histogram(x, bins=int(round(bins[i])), weights=weights,
                                range=range[i])
            n = _gf(n, smooth1d)
            x0 = _np.array(zip(b[:-1], b[1:])).flatten()
            y0 = _np.array(zip(n, n)).flatten()
            ax.plot(x0, y0, **hist_kwargs)

        if truths is not None and truths[i] is not None:
            ax.axvline(truths[i], color=truth_color)

        # Plot quantiles if wanted.
        if len(quantiles) > 0:
            qts = _np.array(quantiles) + qtdiff[i]
            qts = _np.where(qts <= 0, 0.01, qts)
            qts = _np.where(qts >= 1, 0.99, qts)
            qts = list(qts)
            qvalues = quantile(x, qts, weights=weights)
            for q in qvalues:
                ax.axvline(q, ls="dashed", color=color)

            if verbose:
                print("Quantiles:")
                print([item for item in zip(qts, qvalues)])                

        if show_titles:
            # Compute the quantiles for the title. This might redo
            # unneeded computation but who cares.
            qts = _np.array(quantiles) + qtdiff[i]
            qts = _np.where(qts <= 0, 0.01, qts)
            qts = _np.where(qts >= 1, 0.99, qts)
            qts = list(qts)
            q_16, q_50, q_84 = quantile(x, qts, weights=weights)             
            q_m, q_p = q_50 - q_16, q_84 - q_50

            # Format the quantile display.
            if q_50 < 1e6:
                fmt = "{{0:{0}}}".format(title_fmt).format
                title = r"${{{0}}}_{{-{1}}}^{{+{2}}}$"
                title = title.format(fmt(q_50), fmt(q_m), fmt(q_p))
            else:
                expn = int(_np.log10(q_50))
                fmt = "{{0:{0}}}".format(title_fmt).format
                title = r"${{{0}}}_{{-{1}}}^{{+{2}}}e{3}$"
                title = title.format(fmt(q_50 * 10**-expn), fmt(q_m * 
                    10**-expn), fmt(q_p * 10**-expn), expn)

            # Add in the column name if it's given.
            if labels is not None:
                if len(labels) <= i:
                    labels = list(labels)+['']
                j = labels[i].find(' ')
                if j == -1:
                    j = len(labels[i])
                title = "{0} = {1}".format(labels[i][:j], title)

            # Add the title to the axis.
            ax.set_title(title, **title_kwargs)

        # Set up the axes.
        ax.set_xlim(range[i])
        if scale_hist:
            maxn = _np.max(n)
            ax.set_ylim(-0.1 * maxn, 1.1 * maxn)
        else:
            ax.set_ylim(0, 1.1 * _np.max(n))
        ax.set_yticklabels([])
        ax.xaxis.set_major_locator(_MaxNLoc(max_n_ticks, prune="lower"))

        if i < K - 1:
            if top_ticks:
                ax.xaxis.set_ticks_position("top")
                [l.set_rotation(45) for l in ax.get_xticklabels()]
            else:
                ax.set_xticklabels([])
        else:
            [l.set_rotation(45) for l in ax.get_xticklabels()]
            if labels is not None:
                ax.set_xlabel(labels[i], **label_kwargs)
                ax.xaxis.set_label_coords(0.5, -0.3)

        for j, y in enumerate(xs):
            if _np.shape(xs)[0] == 1:
                ax = axes
            else:
                ax = axes[i, j]
            if j > i:
                ax.set_visible(False)
                ax.set_frame_on(False)
                continue
            elif j == i:
                continue

            # Deal with masked arrays.
            if hasattr(y, "compressed"):
                y = y.compressed()
            hist2d(y, x, ax=ax, range=[range[j], range[i]], weights=weights,
                   color=color, smooth=smooth, bins=[bins[j], bins[i]], 
                   **hist2d_kwargs)

            if truths is not None:
                if truths[i] is not None and truths[j] is not None:
                    ax.plot(truths[j], truths[i], "s", color=truth_color)
                if truths[j] is not None:
                    ax.axvline(truths[j], color=truth_color)
                if truths[i] is not None:
                    ax.axhline(truths[i], color=truth_color)

            ax.xaxis.set_major_locator(_MaxNLoc(max_n_ticks, prune="lower"))
            ax.yaxis.set_major_locator(_MaxNLoc(max_n_ticks, prune="lower"))

            if i < K - 1:
                ax.set_xticklabels([])
            else:
                [l.set_rotation(45) for l in ax.get_xticklabels()]
                if labels is not None:
                    ax.set_xlabel(labels[j], **label_kwargs)
                    ax.xaxis.set_label_coords(0.5, -0.3)

            if j > 0:
                ax.set_yticklabels([])
            else:
                [l.set_rotation(45) for l in ax.get_yticklabels()]
                if labels is not None:
                    ax.set_ylabel(labels[i], **label_kwargs)
                    ax.yaxis.set_label_coords(-0.3, 0.5)

    return fig


def quantile(x, q, weights=None):
    """
    Like numpy.percentile, but:

    * Values of q are quantiles [0., 1.] rather than percentiles [0., 100.]
    * scalar q not supported (q must be iterable)
    * optional weights on x

    """
    if weights is None:
        return _np.percentile(x, [100. * qi for qi in q])
    else:
        idx = _np.argsort(x)
        xsorted = x[idx]
        cdf = _np.add.accumulate(weights[idx])
        cdf /= cdf[-1]
        return _np.interp(q, cdf, xsorted).tolist()


def hist2d(x, y, bins=20, range=None, weights=None, levels=None, smooth=None,
           ax=None, color=None, plot_datapoints=True, plot_density=True,
           plot_contours=True, fill_contours=False,
           contour_kwargs=None, contourf_kwargs=None, data_kwargs=None,
           **kwargs):
    """
    Plot a 2-D histogram of samples.

    """
    if ax is None:
        ax = _plt.gca()

    # Set the default range based on the data range if not provided.
    if range is None:
        if "extent" in kwargs:
            _warn.warn("Deprecated keyword argument 'extent'. "
                "Use 'range' instead.")
            range = kwargs["extent"]
        else:
            range = [[x.min(), x.max()], [y.min(), y.max()]]

    # Set up the default plotting arguments.
    # if color is None:
        # color = "k"
# 
    # Choose the default "sigma" contour levels.
    # if levels is None:
        # levels = 1.0 - _np.exp(-0.5 * _np.arange(0.5, 2.1, 0.5) ** 2)
# 
    # This is the color map for the density plot, over-plotted to indicate the
    # density of the points near the center.
    # density_cmap = LinearSegmentedColormap.from_list(
        # "density_cmap", [color, (1, 1, 1, 0)])
# 
    # This color map is used to hide the points at the high density areas.
    # white_cmap = LinearSegmentedColormap.from_list(
        # "white_cmap", [(1, 1, 1), (1, 1, 1)], N=2)
# 
    # This "color map" is the list of colors for the contour levels if the
    # contours are filled.
    # rgba_color = colorConverter.to_rgba(color)
    # contour_cmap = [rgba_color] + [list(rgba_color) for l in levels]
    # for i, l in enumerate(levels):
        # contour_cmap[i+1][-1] *= float(len(levels) - i) / (len(levels)+1)
# 
    # We'll make the 2D histogram to directly estimate the density.
    try:
        H, X, Y = _np.histogram2d(x.flatten(), y.flatten(), bins=bins,
                                 range=range, weights=weights)
    except ValueError:
        raise ValueError("It looks like at least one of your sample columns "
                         "have no dynamic range. You could try using the "
                         "'range' argument.")
# 
    # if smooth is not None:
        # if gaussian_filter is None:
            # raise ImportError("Please install scipy for smoothing")
        # H = _gf(H, smooth)
# 
    # Compute the density levels.
    # Hflat = H.flatten()
    # inds = _np.argsort(Hflat)[::-1]
    # Hflat = Hflat[inds]
    # sm = _np.cumsum(Hflat)
    # sm /= sm[-1]
    # V = _np.empty(len(levels))
    # for i, v0 in enumerate(levels):
        # try:
            # V[i] = Hflat[sm <= v0][-1]
        # except:
            # V[i] = Hflat[0]
# 
    # Compute the bin centers.
    # X1, Y1 = 0.5 * (X[1:] + X[:-1]), 0.5 * (Y[1:] + Y[:-1])
# 
    # Extend the array for the sake of the contours at the plot edges.
    # H2 = H.min() + _np.zeros((H.shape[0] + 4, H.shape[1] + 4))
    # H2[2:-2, 2:-2] = H
    # H2[2:-2, 1] = H[:, 0]
    # H2[2:-2, -2] = H[:, -1]
    # H2[1, 2:-2] = H[0]
    # H2[-2, 2:-2] = H[-1]
    # H2[1, 1] = H[0, 0]
    # H2[1, -2] = H[0, -1]
    # H2[-2, 1] = H[-1, 0]
    # H2[-2, -2] = H[-1, -1]
    # X2 = _np.concatenate([
        # X1[0] + _np.array([-2, -1]) * _np.diff(X1[:2]),
        # X1,
        # X1[-1] + _np.array([1, 2]) * _np.diff(X1[-2:]),
    # ])
    # Y2 = _np.concatenate([
        # Y1[0] + _np.array([-2, -1]) * _np.diff(Y1[:2]),
        # Y1,
        # Y1[-1] + _np.array([1, 2]) * _np.diff(Y1[-2:]),
    # ])
# 
    # if plot_datapoints:
        # if data_kwargs is None:
            # data_kwargs = dict()
        # data_kwargs["color"] = data_kwargs.get("color", color)
        # data_kwargs["ms"] = data_kwargs.get("ms", 2.0)
        # data_kwargs["mec"] = data_kwargs.get("mec", "none")
        # data_kwargs["alpha"] = data_kwargs.get("alpha", 0.1)
        # ax.plot(x, y, "o", zorder=-1, rasterized=True, **data_kwargs)
# 
    # Plot the base fill to hide the densest data points.
    # if plot_contours or plot_density:
        # ax.contourf(X2, Y2, H2.T, [V[-1], H.max()],
                    # cmap=white_cmap, antialiased=False)
# 
    # if plot_contours and fill_contours:
        # if contourf_kwargs is None:
            # contourf_kwargs = dict()
        # contourf_kwargs["colors"] = contourf_kwargs.get("colors", contour_
            # cmap)
        # contourf_kwargs["antialiased"] = contourf_kwargs.get("antialiased",
            # False)
        # ax.contourf(X2, Y2, H2.T, _np.concatenate([[H.max()], V, [0]]),
                    # **contourf_kwargs)
# 
    # Plot the density map. This can't be plotted at the same time as the
    # contour fills.
    # elif plot_density:
        # ax.pcolor(X, Y, H.max() - H.T, cmap=density_cmap)
    if plot_density:
        ax.pcolor(X, Y, H.max() - H.T, cmap=_cm.get_cmap('gist_heat'))
# 
    # Plot the contour edge colors.
    # if plot_contours:
        # if contour_kwargs is None:
            # contour_kwargs = dict()
        # contour_kwargs["colors"] = contour_kwargs.get("colors", color)
        # ax.contour(X2, Y2, H2.T, V, **contour_kwargs)

    ax.set_xlim(range[0])
    ax.set_ylim(range[1])
    return

# Main
if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-

""" Program to export as figure line profiles from standard FITS spectra.
"""

import sys
from glob import glob
import pyhdust.spectools as spt
import pyhdust.phc as phc
from argparse import ArgumentParser
import os
import matplotlib.pyplot as plt

__version__ = "0.91"
__author__ = "Daniel Moser"
__email__ = "dmfaes@gmail.com"


class MyParser(ArgumentParser): 
    def error(self, message):
        sys.stderr.write('# ERROR! %s\n' % message)
        self.print_help()
        sys.exit(2)

parser = MyParser(description=__doc__)
parser.add_argument('--version', action='version', 
    version='%(prog)s {0}'.format(__version__))
parser.add_argument("INPUT", help=("String to retrive the spectrum(a) "
    "filename(s) (wildcards accepted wrapped with \"\")"), type=str)
parser.add_argument("-l", "--line", action="store", dest="line", 
    help=("Wavelength of the reference line [default: %(default)s]"), 
    type=float, default=6562.79)
parser.add_argument("-w", "--hwidth", action="store", dest="hwidth", 
    help=("Half-width of the line profile (in km/s) [default: %(default)s]"), 
    type=float, default=1000.)

# group2 = parser.add_argument_group('other arguments (overwrite power)')
# group2.add_argument("-r", "--remove", action="store_true", dest="rm", 
#     help=("If this flag is enabled, it restores the spectrum(a) to the "
#         "original value (i.e., VELSHIFT = 0)"), default=False)
# group2.add_argument("-f", "--force-dv", action="store", dest="vs", 
#     help=("Force VELSHIFT to the non-zero VS value (in km/s) "
#         "[default: None]"), type=float, default=0.)

args = parser.parse_args()


if __name__ == '__main__':

    lbc = args.line
    linput = args.INPUT
    hwidth = args.hwidth

    plev = 0.
    lfiles = glob(linput)
    lfiles.sort()
    if len(lfiles):
        fig, ax = plt.subplots(figsize=(8, 6*(1+len(lfiles)/25.)))

    for fitsfile in lfiles:
        sdata = spt.loadfits(fitsfile)
        vl, fx = spt.lineProf(sdata[0], sdata[1], lbc=lbc, hwidth=hwidth)
        ax.plot(vl, fx+plev)
        ax.text(0.95*hwidth, 1.+plev+0.05/2, os.path.splitext(fitsfile)[0], 
            horizontalalignment='right',  # transform=ax.transAxes,
        verticalalignment='center', fontsize=8)
        if plev == 0:
            ymin = min(fx)
        plev += 0.04

    if len(glob(linput)) == 0:
        print('# ERROR! {0} files were not found!'.format(linput))
    else:
        # ax.legend(loc='best', fancybox=True, framealpha=0.5, fontsize=8, 
        #         labelspacing=0.05)
        ax.set_ylim([ymin, plev+max(fx)])
        ax.set_xlim([-hwidth, hwidth])
        phc.savefig(fig, figname='plot_spec_'+phc.dtflag())



# -*- coding:utf-8 -*-

"""PyHdust *hdrpil* module: third-part functions for HDR process of images

import pyhdust.hdrpil asd hdr
proj = hdr.hdr(case='img04', img_type='png', cur_dir='./')
proj.get_hdr(strength=[1.],naturalness=[1.0,1.1,1.2,1.5])

:license: ?
"""
from __future__ import print_function
import os
from copy import copy
import warnings as _warn

try:
    from PIL import Image
    import pylab
    pylab_loaded = 1
except ImportError:
    pylab_loaded = 0
    _warn.warn('matplotlib+pylab and/or PIL/Pillow not installed!!!')

__author__ = "bpowah"
__email__ = "bpowah@gmail.com"


class hdr:

    """
    a collection of images to merege into HDR

    blend an arbitrary number of photos into a single HDR image
    or several images with various combinations of HDR parameters

    it is assumed that project folders contain all the images you want to merge
    case = folder name
    cur_dir = folder where the project folders are located
    images are auto-sorted by degree of exposure (image brightness)
    """

    def get_imgs(self):
        """
        load a list of images from folder
        sort them from lightest to darkest
        """
        drks = []
        imgs = [Image.open(os.path.join(self.indir, fn)) for fn in self.fns]
        for img in imgs:
            samp = img.resize((50, 50))  # crop(box=((10,10,20,20)))
            drks.append(sum(samp.getdata(band=0)))
        if self.resize is not None:
            newsize = tuple([int(x * self.resize) for x in img.size])
            imgs = [img.resize(newsize) for img in imgs]

        newdrks = sorted(drks)
        idxs = [drks.index(x) for x in newdrks]
        imgs = [imgs[x] for x in idxs]
        self.fns = [self.fns[x] for x in idxs]

        self.drks = drks
        self.idxs = idxs

        print('got', len(imgs), 'images')
        return imgs

    def __init__(
        self,
        case='',
        resize=None,
        img_type='.jpg',
        cur_dir=r'./' ):
        """
        load a project
        all images of [img_type] are loaded from folder [case]
        and resized by a factor [resize]
        """
        if not case:
            raise ValueError('case is required')
        self.resize = resize
        self.ext = img_type.strip('.')
        self.case = case
        self.cur_dir = os.path.normpath(cur_dir)
        self.indir = os.path.join(self.cur_dir, case)
        self.outdir = os.path.join(self.indir, 'out')

        if not os.path.isdir(self.outdir):
            os.makedirs(self.outdir)
        f = lambda x: x.upper().endswith(self.ext.upper())
        self.fns = filter(f, os.listdir(self.indir))
        self.imgs = self.get_imgs()
        self.fin_set = {}

    def get_masks(self, imgs):
        """
        create a set of masks from a list of images
        (one mask for every adjacent pair of images
        """
        masks = []
        mask_ct = len(imgs) - 1
        imgs = [self.bal(img.convert(mode='L')) for img in imgs]
        for i in range(mask_ct):
            blend_fraction = .5  # 1. - (float(i)+.5)/float(mask_ct)
            m = Image.blend(imgs[i], imgs[i + 1], blend_fraction)
            masks.append(m)
            # print blend_fraction
        print('blending using', mask_ct, 'masks')
        return masks

    def lev(self, im):
        # im = ImageEnhance.Brightness(im).enhance(self.bri)
        # im = ImageEnhance.Contrast(im).enhance(self.con)
        return self.bal(im, self.str)

    def bal(self, im):
        """
        adjust the balance of the mask
        (re-distribute the histogram so that there are more
        extreme blacks and whites)
        like increasing the contrast, but without clipping
        and maintains overall average image brightness
        """
        h = im.histogram()
        ln = range(len(h))
        up = [sum(h[0: i]) for i in ln]
        lo = [sum(h[i:-1]) for i in ln]
        ct = sum(h)
        st = int(self.cur_str * 255.)

        lut = [i + st * up[i] * lo[i] * (up[i] - lo[i]) / ct**3 for i in ln]
        for i in ln:
            if lut[i] < 1:
                lut[i] = 1
            if lut[i] > 255:
                lut[i] = 255
        return im.point(lut)

    def save_im(self, im, name):
        """
        save an image
        """
        print('saving', name)
        im.save(os.path.join(self.outdir, name + '.jpg'), format='JPEG')

        if pylab_loaded:
            pylab.cla()
            h = im.histogram()
            rgb = h[0:256], h[256:256 * 2], h[256 * 2:]
            x = range(256)
            for b in rgb:
                pylab.plot(x, b)
            pylab.xlim(xmax=255)
            pylab.savefig(os.path.join(self.outdir, name + '_histogram.png'))

    def merge(self, imgs):
        """
        combine a set images into a smaller set by combinding all
        adjacent images
        """
        masks = self.get_masks(imgs)
        imx = lambda i: Image.composite(imgs[i], imgs[i + 1], masks[i])
        return [imx(i) for i in range(len(masks))]

    def merge_all(self, imgs):
        """
        iteratively merge a set of images until only one remains
        """
        while len(imgs) > 1:
            imgs = self.merge(imgs)
        return imgs[0]

    def get_hdr(self, strength=[0.0, 1.0, 2.0], naturalness=[0.8, 0.9, 1.0]):
        """
        process the hdr image(s)
        strength - a list or a float that defines how strong the hdr
                   effect should be
                 - a value of zero will combine images by using a
                   greyscale image average
                 - a value greater than zero will use higher contrast
                   versions of those greyscale images
        naturalness- values between zero and one
                 - zero will be a very high-contrast image
                 - 1.0 will be a very flat image
                 - 0.7 to 0.9 tend to give the best results
        """
        contrast = 1.0
        brightness = 1.0
        self.con = contrast  # not used
        self.bri = brightness  # not used
        self.nat = self.to_list(naturalness)
        self.str = self.to_list(strength)
        self.fin_set = {}

        for s in self.str:
            self.cur_str = s
            print('getting saturation image, strength', str(s))
            imgs = copy(self.imgs)
            sat_img = self.merge_all(imgs)

            print('getting contrast image')
            imgs.reverse()
            con_img = self.merge_all(imgs)

            self.final_blend(con_img, sat_img)

        self.print_set(self.fin_set)
        ori_set = dict(
            ('orig_' + str(i), self.imgs[i]) for i in range(len(self.imgs)))
        self.print_set(ori_set)
        print(self.indir)

    def final_blend(self, im1, im2):
        """
        combines a saturated image with a contrast image
        and puts them in a dictionary of completed images
        """
        for nat in self.nat:
            n_s = '_nat' + str(round(nat, 2))
            s_s = '_str' + str(round(self.cur_str, 2))
            s = self.case + s_s + n_s
            self.fin_set[s] = Image.blend(im1, im2, nat)

    def print_set(self, im_dict):
        """
        print all rendered images
        """
        for k in im_dict.keys():
            self.save_im(im_dict[k], k)

    def to_list(self, val):
        # if type(val) != type(list()):
        if not isinstance(val, list):
            val = [float(val)]
        else:
            val = [float(v) for v in val]
        return val

# MAIN ###
if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-

import os
import sys


def check_py_ver():
    if (sys.version_info[0] > 2):
        return
    elif (sys.version_info[1] >= 7):
        return
    else: 
        _ = raw_input('Press Ctlr+Z to abort or ENTER to continue... ')
        return


print('# Warning! PyHdust in only supported for Python 2.7+\n'
    'Press Ctlr+Z to abort or ENTER to continue... ')
check_py_ver()

try:
    from setuptools import setup, find_packages
    errimport = False
except ImportError:
    from distutils.core import setup
    from glob import glob
    errimport = True

    def is_package(path):
        return ( os.path.isdir(path) and os.path.isfile(os.path.join(path, 
            '__init__.py')) )

    def find_packages(path=".", base="", exclude=[]):
        """ Find all packages in path """
        packages = {}
        lexc = []
        for item in exclude:
            lexc.extend(glob(item))
        for item in os.listdir(path):
            dir = os.path.join(path, item)
            if is_package( dir ) and item not in lexc:
                if base:
                    module_name = "%(base)s.%(item)s" % vars()
                else:
                    module_name = item
                packages[module_name] = dir
                packages.update(find_packages(dir, module_name))
        return packages


def rd(filename):
    f = open(filename)
    r = f.read()
    f.close()
    return r


def recfiles(chdir, path):
    out = []
    opath = os.getcwd()
    os.chdir(chdir)
    for root, dirs, fnames in os.walk(path):
        for f in fnames:
            out.append(os.path.join(root, f))
    os.chdir(opath)
    return out


if sys.argv[-1] == "publish":
    os.system("python setup.py sdist upload")
    sys.exit()

setup(name='pyhdust',
    version='1.3.2',
    description=('Analysis tools for multi-technique astronomical data and '
        'hdust models'),
    url='http://pyhdust.readthedocs.io',
    author='Daniel M. Faes',
    author_email='dmfaes@gmail.com',
    license='GNU GPLv3.0',      
    # packages=['pyhdust','pyhdust_refs'],
    scripts=[os.path.join('scripts', f) for f in os.listdir('scripts') if 
        os.path.splitext(f)[1] == '.py'],
    packages=find_packages(exclude=['build', 'docs', '*egg*', 'dist', 
        'scripts']),
    # include=recfiles('refs'),
    package_data={'pyhdust': recfiles('pyhdust', 'refs') + 
        ['LICENSE', 'README.rst']},
    include_package_data=True,
    # data_files = [('refs/*', 'stmodels/*')],
    # package_dir = {'../'},
    zip_safe=False,
    install_requires=['numpy >= 1.9.0', 'six'],
    long_description=rd('README.rst'),
    classifiers=[
        # "Development Status :: 4 - Beta",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later" + \
        " (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Physics"
    ],
    keywords=['astronomy', 'astrophysics', 'science', 'hdust', 'Be stars',
        'spectroscopy', 'polarimetry', 'interferometry', 'radiative transfer',
        'optical-interferometry'],
)

if errimport:
    print('# You don\'t have "setuptools" installed!')
    print('# Because of this, you MAY need to copy the package data: \n')
    print('# Warning! The path can change according to your system')
    print('$ cp -r -f pyhdust ~/.local/lib/python2.7/site-packages/')



# -*- coding:utf-8 -*-

"""PyHdust *beatlas* module: BeAtlas specific variables and functions.

Module contains:
- BAstar class
- BAmod class

:license: GNU GPL v3.0 https://github.com/danmoser/pyhdust/blob/master/LICENSE
"""
from __future__ import print_function
import re as _re
import os as _os
import numpy as _np
import struct as _struct
from glob import glob as _glob
from itertools import product as _product
import pyhdust.phc as _phc
import pyhdust as _hdt
import warnings as _warn

try: 
    from scipy.interpolate import griddata as _griddata
except ImportError:
    _warn.warn('# scipy module not installed!')

__author__ = "Daniel Moser"
__email__ = "dmfaes@gmail.com"


class BAstar(object):

    """ BeAtlas source star filename structure.

    The filename must follow this structure (the sequence in filename is not 
    importante):
    - keys: ['M', 'ob', 'H', 'Z', 'b']
    - last-key (no identifier): 'Ell'

    See BAmod, phc.keys_values.
    """

    def __init__(self, f0):
        vals = _phc.keys_values(['M', 'ob', 'H', 'Z', 'b'], f0)
        self.M, self.ob, self.H, self.Z, self.beta = vals
        self.shape = f0.split('_')[-1].replace('.txt', '')
        self._f0 = f0

    def __repr__(self):
        return self._f0


class BAmod(BAstar):

    """ BeAtlas disk model filename structure.

    It could be f0.split('_'), but the f0.find('_X') way was chosen.

    See that the parameters sequence is not important for this reading (this
    may not be the case of other routines). And, by definition, the source star
    has a specific name added at the end of disk model name, starting with
    'Be_'. """

    def __init__(self, f0):
        """ Class initialiser """
        BAstar.__init__(self, f0[f0.find('Be_'):])
        self.param = False
        if f0.find('_PL') > -1:
            self.param = True
            self.n = _phc.keys_values(['PLn'], f0)[0]
        self.sig, self.h, self.Rd = _phc.keys_values(['sig', 'h', 'Rd'], f0)

    def build(self, ctrlarr, listpars):
        """ Set full list of parameters. """
        for i in range(len(ctrlarr)):
            if i == 0:
                self.M = _phc.find_nearest(listpars[i], ctrlarr[i])
            if i == 1:
                self.ob = _phc.find_nearest(listpars[i], ctrlarr[i])
            if i == 2:
                self.Z = _phc.find_nearest(listpars[i], ctrlarr[i])
            if i == 3:
                self.H = _phc.find_nearest(listpars[i], ctrlarr[i])
            if i == 4:
                self.sig = _phc.find_nearest(listpars[i], ctrlarr[i])
            if i == 5:
                self.Rd = _phc.find_nearest(listpars[i], ctrlarr[i])
            if i == 6:
                self.h = _phc.find_nearest(listpars[i], ctrlarr[i])
            if len(listpars) == 9:
                if i == 7:
                    # print(_phc.find_nearest(listpars[i], ctrlarr[i]))
                    self.n = _phc.find_nearest(listpars[i], ctrlarr[i])
                    self.param = True
                if i == 8:
                    self.cosi = _phc.find_nearest(listpars[i], ctrlarr[i])
            else:
                if i == 7:
                    self.cosi = _phc.find_nearest(listpars[i], ctrlarr[i])

    def getidx(self, minfo):
        """ Find index of current model in minfo array. """
        if len(minfo[0]) == 9:
            self.idx = (minfo[:, 0] == self.M) & (minfo[:, 1] == self.ob) &\
                (minfo[:, 2] == self.Z) & (minfo[:, 3] == self.H) &\
                (minfo[:, 4] == self.sig) & (minfo[:, 5] == self.Rd) &\
                (minfo[:, 6] == self.h) & (minfo[:, 7] == self.n) &\
                (minfo[:, -1] == self.cosi)
        else:
            self.idx = (minfo[:, 0] == self.M) & (minfo[:, 1] == self.ob) &\
                (minfo[:, 2] == self.Z) & (minfo[:, 3] == self.H) &\
                (minfo[:, 4] == self.sig) & (minfo[:, 5] == self.Rd) &\
                (minfo[:, 6] == self.h) & (minfo[:, -1] == self.cosi)
        return self.idx


# Only for H=0.30
vrots = [
    [259.759, 354.834, 417.792, 464.549, 483.847],
    [252.050, 346.163, 406.388, 449.818, 468.126],
    [245.127, 336.834, 399.983, 448.076, 467.806],
    [239.522, 329.496, 388.734, 432.532, 450.806],
    [234.301, 321.139, 379.297, 423.241, 441.122],
    [228.538, 313.797, 370.343, 412.488, 429.914],
    [219.126, 299.656, 354.547, 395.821, 413.008],
    [211.544, 288.840, 341.081, 380.426, 396.978],
    [203.438, 279.328, 328.666, 365.697, 380.660],
    [197.823, 268.964, 316.901, 353.568, 368.506],
    [192.620, 262.688, 308.208, 341.963, 356.410],
    [187.003, 255.125, 299.737, 332.511, 346.043]]

obs = [1.1, 1.2, 1.3, 1.4, 1.45]
ms = [14.6, 12.5, 10.8, 9.6, 8.6, 7.7, 6.4, 5.5, 4.8, 4.2, 3.8, 3.4]
xc = [0.08, 0.30, 0.42, 0.54, 0.64, 0.77]
ns = [3.0, 3.5, 4.0, 4.5]
nia = [0.5]
z = [0.014]
h = [72]
Rd = [50.0]
sig0 = _np.logspace(_np.log10(0.02), _np.log10(4.0), 7)

# Only for H=0.30
Tp11 = _np.array([28905.8, 26945.8, 25085.2, 23629.3, 22296.1, 20919.7,
18739.3, 17063.8, 15587.7, 14300.3, 13329.9, 12307.1])
Ms = _np.array([14.6, 12.5, 10.8, 9.6, 8.6, 7.7, 6.4, 5.5, 4.8, 4.2, 3.8, 3.4],
dtype=str)
Sig0 = ['{0:.2f}'.format(x) for x in sig0]


def rmMods(modn, Ms, clusters=['job']):
    """
    Remove the *.inp models of models `modn` according to the list structure
    below.

    | Masses list ans sig0 POSITION do be excluded
    | Ms = [
    | ['14.6', [0]],
    | ['12.5', [0,-1]],
    | ['10.8', [0,-1]],
    | ['09.6',  [0,-2,-1]],
    | ['08.6',  [0,-2,-1]],
    | ['07.7',  [0,-2,-1]],
    | ['06.4',  [0,-3,-2,-1]],
    | ['05.5',  [0,-3,-2,-1]],
    | ['04.8',  [-4,-3,-2,-1]],
    | ['04.2',  [-4,-3,-2,-1]],
    | ['03.8',  [-4,-3,-2,-1]],
    | ['03.4',  [-4,-3,-2,-1]],]

    INPUT: string, structured list

    OUTPUT: *files removed
    """
    # Create sig0 list
    sig0s = Sig0
    project = _phc.trimpathname(_os.getcwd())[1]
    for cl in clusters:
        file = open('{0}s/{0}s_{1}_mod{2}.sh'.format(cl, project, modn))
        lines = file.readlines()
        file.close()
        for item in Ms:
            M = item[0]
            exsig = item[1]
            for rm in exsig:
                _os.system('rm mod{0}/mod{0}*_sig{1}*_M{2}*.inp'.format(modn,
                sig0s[rm], M))
                print('# Deleted mod{0}/mod{0}*_sig{1}*_M{2}*.inp'.format(modn,
                sig0s[rm], M))
                _os.system('rm {3}s/mod{0}*_sig{1}*_M{2}*.{3}'.format(modn,
                sig0s[rm], M, cl))
                lines = [line for line in lines if (line.find('_sig{0}'.format(
                    sig0s[rm])) == -1 or line.find('_M{0}'.format(M)) == -1)]
        file = open('{0}s/{0}s_{1}_mod{2}.sh'.format(cl, project, modn), 'w')
        file.writelines(lines)
        file.close()
    # End prog
    return


def fsedList(fsedlist, param=True):
    """ Return the total of models and the parameters values in the fullsed list.

    The len of fsedlist is 9 (param=True) for the parametric case and 8
    to the VDD-ST one.

    The sequence is: M, ob(W), Z, H, sig, Rd, h, *n*, cos(i).

    It is assumed that all models have the same `observers` configuration."""
    nq = 9
    if not param:
        nq = 8
    listpar = [[] for i in range(nq)]
    nm = 0
    for sed in fsedlist:
        mod = BAmod(sed)
        if mod.param == param:
            nm += 1
            if mod.M not in listpar[0]:
                listpar[0].append(mod.M)
            if mod.ob not in listpar[1]:
                listpar[1].append(mod.ob)
            if mod.Z not in listpar[2]:
                listpar[2].append(mod.Z)
            if mod.H not in listpar[3]:
                listpar[3].append(mod.H)
            if mod.sig not in listpar[4]:
                listpar[4].append(mod.sig)
            if mod.Rd not in listpar[5]:
                listpar[5].append(mod.Rd)
            if mod.h not in listpar[6]:
                listpar[6].append(mod.h)
            if param:
                if mod.n not in listpar[7]:
                    listpar[7].append(mod.n)
            if listpar[-1] == []:
                sed2data = _hdt.readfullsed2(sed)
                listpar[-1] = list(sed2data[:, 0, 0])
    #
    for vals in listpar:
        vals.sort()
    return nm * len(listpar[-1]), listpar


def createBAsed(fsedlist, xdrpath, lbdarr, param=True, savetxt=False,
    ignorelum=False, pol=False, saveextra=None):
    """ Create the BeAtlas SED XDR release.

    WARNING: The file names must be in this format: 
    `mod01_PLn3.5_sig0.00_h072_Rd000.0_Be_M14.60_ob1.45_H0.77_Z0.014_bE_Ell`

    | The file structure:
    | -n_quantities, n_lbd, n_models,
    | -n_qt_vals1, n_qt_vals2, .. n_qt_valsn
    | -quantities values =  M, ob(W), Z, H, sig, Rd, h, *n*, cos(i).
    | -(Unique) lbd array
    | -Loop:
    |   *model values
    |   *model SED

    | Definitions:
    | -photospheric models: sig0 = 0.00
    | -Parametric disk model default (`param` == True)
    | -VDD-ST models: n excluded (alpha and R0 fixed. Confirm?)
    | -The flux will be given in ergs/s/um2/um. If ignorelum==True, the usual
    |   F_lbda/F_bol unit will be given.

    Since the grid is not symmetric, there is no index to jump directly to the
    desired model. So the suggestion is to use the index matrix, or read the
    file line by line until find the model (if exists).

    :Example: 

    def genxdr(xdrname='PL.xdr', param=True, pol=False):
        fs2l = glob('fullsed/*.sed2')
        print('# Using {0} as reference!'.format(fs2l[0]))
        lbdarr = hdt.readfullsed2(fs2l[0])
        lbdarr = lbdarr[0, :, 2]
        nm, listpar = bat.fsedList(fs2l)
        bat.createBAsed(fs2l, xdrname, lbdarr, param=param, savetxt=False, 
            pol=pol, saveextra=xdrname.replace('xdr', 'txt'))
        return

    genxdr(xdrname='Yudin_PL.xdr')
    """
    fsedlist.sort()
    nq = 9
    if not param:
        nq = 8
    nm, listpar = fsedList(fsedlist, param=param)
    header2 = []
    for vals in listpar:
        header2 += [len(vals)]
    nlb = len(lbdarr)
    header1 = [nq, nlb, nm]
    models = _np.zeros((nm, nlb))
    minfo = _np.zeros((nm, nq))
    k = 0
    iflx = 3
    if pol:
        iflx = 7
    for i in range(len(fsedlist)):
        mod = BAmod(fsedlist[i])
        # Select only `param` matching cases:
        if mod.param == param:
            sed2data = _hdt.readfullsed2(fsedlist[i])
            iL = 1.
            dist = 1/_np.sqrt(4 * _np.pi)
            if not ignorelum and not pol:
                j = fsedlist[i].find('fullsed_mod')
                # modn = fsedlist[i][j + 11:j + 13]
                modn = _re.match(r'.*mod(\d+)_', fsedlist[i]).group(1)
                log = fsedlist[i].replace('fullsed_mod', '../mod{0}/mod'.
                    format(modn)).replace('.sed2', '.log')
                if not _os.path.exists(log):
                    log = _glob(log.replace('../mod{0}/mod'.format(modn),
                    '../mod{0}/*mod'.format(modn)))
                    if len(log) >= 1:
                        log = log[0]
                    else:
                        raise LookupError('# No log file found for {0}'.
                            format(fsedlist[i]))
                f0 = open(log)
                lines = f0.readlines()
                f0.close()
                iL = _phc.fltTxtOccur('L =', lines) * _phc.Lsun.cgs
                if saveextra is not None:
                    R_pole = _phc.fltTxtOccur('R_pole =', lines)
                    Vrot = _phc.fltTxtOccur('Vrot', lines)
                    f0 = open(saveextra, 'a')
                    f0.writelines('{0}\t{1}\t{2}\n'.format(R_pole, Vrot, iL))
                    f0.close()
                dist = 10. * _phc.pc.cgs
            for j in range(header2[-1]):
                #  M, ob(W), Z, H, sig, Rd, h, *n*, cos(i).
                if param:
                    minfo[k * header2[-1] + j] = _np.array([ mod.M, mod.ob, 
                    mod.Z, mod.H, mod.sig, mod.Rd, mod.h, mod.n, 
                    listpar[-1][j] ]).astype(float)
                else:
                    minfo[k * header2[-1] + j] = _np.array([ mod.M, mod.ob, 
                    mod.Z, mod.H, mod.sig, mod.Rd, mod.h,
                    listpar[-1][j] ]).astype(float)
                if len(sed2data[j, :, 2]) != nlb:
                    models[k * header2[-1] + j] = _np.interp(lbdarr, 
                        sed2data[j, :, 2], sed2data[j, :, iflx]) * iL / 4 / \
                        _np.pi / dist**2
                else:
                    models[k * header2[-1] + j] = sed2data[j, :, iflx] * \
                        iL / 4 / _np.pi / dist**2
                if _np.sum(_np.isnan(models[k * header2[-1] + j])) > 0:
                    nans, x = _phc.nan_helper(models[k * header2[-1] + j]) 
                    models[k * header2[-1] + j][nans] = _np.interp(x(nans), 
                        x(~nans), models[k * header2[-1] + j][~nans])
            k += 1
    #
    f0 = open(xdrpath, 'wb')
    stfmt = '>{0}l'.format(3)
    f0.write(_struct.pack(stfmt, *header1))
    stfmt = '>{0}l'.format(nq)
    f0.write(_struct.pack(stfmt, *header2))
    for vals in listpar:
        stfmt = '>{0}f'.format(len(vals))
        f0.write(_struct.pack(stfmt, *_np.array(vals).astype(float)))
    stfmt = '>{0}f'.format(nlb)
    f0.write(_struct.pack(stfmt, *_np.array(lbdarr).astype(float)))
    for i in range(nm):
        stfmt = '>{0}f'.format(nq)
        f0.write(_struct.pack(stfmt, *minfo[i]))
        stfmt = '>{0}f'.format(nlb)
        f0.write(_struct.pack(stfmt, *_np.array(models[i]).astype(float)))
    f0.close()
    print('# XDR file {0} saved!'.format(xdrpath))

    if savetxt:
        f0 = open(xdrpath + '.txt', 'w')
        f0.writelines('{0} \n'.format(header1))
        f0.writelines('{0} \n'.format(header2))
        for vals in listpar:
            f0.writelines('{0} \n'.format(vals))
        f0.writelines('{0} \n'.format(lbdarr))
        for i in range(nm):
            f0.writelines('{0} \n'.format(minfo[i]))
            f0.writelines('{0} \n'.format(models[i]))
        f0.close()
        print('# TXT file {0} saved!'.format(xdrpath + '.txt'))
    return


def createXDRmap(maplist, xdrpath, refclass, lbdlim=None, npix=128):
    """ Is this possible? 

    ``lbdlim``: discard images that are out of this limits (``None`` keeps all;
    units of microns).

    Criteria:
    - all the images are converted the same size in pixels (``npix``).
    - only images with zoom = `renv` are used (checked by `*.log` file)
    - the pixel values are converted to flux units (erg/s/Ang)
    - the pixel scale is converted to length unit (at d = 10 pc)

    Example:
    maspp = maspp * 10/5  # transform the pixel scale for d = 5 pc
    """
    # maspp = milli arcsec per pixel
    raise NotImplementedError("To be done!")
    dval, renv, maspp, cubeimages = range(4)
    listpar = [dval, renv, maspp]
    return listpar, cubeimages


def createXDRsed(fsedlist, xdrpath, refclass, lbdarr, ignorelum=False, 
    pol=False):
    """ Create the generic SED XDR release.

    nob = (individual) number of observers
    listpar = parameters of each model
    nq = number of parameters
    nmod = number of models

    output units: 10**-4 erg/s/Ang/cm2
    """
    if pol:
        ignorelum = True
    fsedlist.sort()
    ifact = 1.
    nq = len(refclass.vdict.keys())+1
    nlbd = len(lbdarr)
    listpar = _np.zeros( nq )
    models = _np.zeros( nlbd )
    for fs in fsedlist:
        print('# Processing {0}'.format(fs))
        m = refclass(fs)
        nob = m.get_nob()
        m.readfs2ob(pol=pol)
        if not ignorelum:
            ifact = m.get_lum() * _phc.Lsun.cgs / (4 * _np.pi * 
                (10 * _phc.pc.cgs)**2) * 1e4 * 1e4
        # 1e4 = from erg/s/um2/um to erg/s/cm$^{2}$/Ang
        # 1e4 = to deliver 10**-4 erg/s/Ang/cm2
        for j in range(nob):
            listpar = _np.vstack(( listpar, [ getattr(m, it) for it in 
                refclass.vdict.keys()]+[ m.obs[j] ] ))
            models = _np.vstack(( models, 
                _np.interp(lbdarr, m.arr_lbd, m.arr_flx[j])*ifact ))

    models = models[1:]
    listpar = listpar[1:]
    nmod = len(models)

    chk = []
    for i in range( nq ):
        if len(_np.unique(listpar[:, i])) == 1:
            chk.append(i)
    if len(chk) > 0:
        listpar = _np.delete(listpar, chk, axis=1)
    nq = len(listpar[0])
    print(nq, _np.shape(listpar), _np.shape(models))
    #
    intervals = _np.zeros(( nq, 2 ))
    intervals[:, 0] = _np.min(listpar, axis=0)
    intervals[:, 1] = _np.max(listpar, axis=0)
    #
    f0 = open(xdrpath, 'wb')
    stfmt = '>{0}l'.format(3)
    f0.write( _struct.pack(stfmt, nq, nlbd, nmod) )
    print(stfmt, nq, nlbd, nmod)
    stfmt = '>{0}f'.format(nq*2)
    f0.write( _struct.pack(stfmt, *intervals.flatten()) )
    stfmt = '>{0}f'.format(nlbd)
    f0.write( _struct.pack(stfmt, *lbdarr))
    stfmt = '>{0}f'.format(nq*nmod)
    f0.write( _struct.pack(stfmt, *listpar.flatten()) )
    stfmt = '>{0}f'.format(nlbd*nmod)
    f0.write( _struct.pack(stfmt, *models.flatten()) )
    f0.close()
    print('# XDR file {0} saved!'.format(xdrpath))
    return


def create_custom_sed(xdrpath, listpar, lbdarr, minfo, models):
    """ Create the generic SED XDR release.

    nob = (individual) number of observers
    listpar = parameters of each model
    nq = number of parameters
    nmod = number of models

    output units: 10**-4 erg/s/Ang/cm2
    """
    nmod = len(minfo)
    nq = len(minfo[0])
    nlbd = len(lbdarr)
    #
    lplen = [len(i) for i in listpar]
    #
    lbdarr = _np.array(lbdarr)
    minfo = _np.array(minfo)
    lbdarr = _np.array(lbdarr)
    f0 = open(xdrpath, 'wb')
    stfmt = '>{0}l'.format(3)
    f0.write( _struct.pack(stfmt, nq, nlbd, nmod) )
    stfmt = '>{0}l'.format(nq)
    f0.write( _struct.pack(stfmt, *_np.array(lplen)) )
    for i in range(nq):
        stfmt = '>{0}f'.format(lplen[i])
        f0.write( _struct.pack(stfmt, *listpar[i].flatten()) )
    stfmt = '>{0}f'.format(nlbd)
    f0.write( _struct.pack(stfmt, *lbdarr))
    stfmt = '>{0}f'.format((nq+nlbd)*nmod)
    tmp = _np.hstack((minfo, models))
    f0.write( _struct.pack(stfmt, *tmp.flatten()) )
    f0.close()
    print('# XDR file {0} saved!'.format(xdrpath))
    return


def readXDRsed(xdrpath, quiet=False):
    """ Doc
    """
    ixdr = 0
    f = open(xdrpath, 'rb').read()
    ixdr, ninfo = _phc.readpck(3, 'l', ixdr, f)
    nq, nlbd, nm = ninfo
    ixdr, intervals = _phc.readpck(nq*2, 'f', ixdr, f)
    ixdr, lbdarr = _phc.readpck(nlbd, 'f', ixdr, f)
    ixdr, listpar = _phc.readpck(nq*nm, 'f', ixdr, f)
    ixdr, models = _phc.readpck(nlbd*nm, 'f', ixdr, f)
    #
    if ixdr == len(f):
        if not quiet:
            print('# XDR {0} completely read!'.format(xdrpath))
    else:
        _warn.warn('# XDR {0} not completely read!\n# length '
            'difference is {1} /4'.format(xdrpath), (len(f)-ixdr) )
    # 
    return ( ninfo, intervals.reshape((nq, 2)), lbdarr, 
        listpar.reshape((nm, nq)), models.reshape((nm, nlbd)) )


def readBAsed(xdrpath, quiet=False):
    """ Read the BeAtlas SED release.

    | Definitions:
    | -photospheric models: sig0 (and other quantities) == 0.00
    | -Parametric disk model default (`param` == True)
    | -VDD-ST models: n excluded (alpha and R0 fixed. Confirm?)
    | -The models flux are given in ergs/s/cm2/um. If ignorelum==True in the
    |   XDR creation, F_lbda/F_bol unit will be given.

    INPUT: xdrpath

    | OUTPUT: listpar, lbdarr, minfo, models 
    | (list of mods parameters, lambda array (um), mods index, mods flux)
    """
    f = open(xdrpath, 'rb').read()
    ixdr = 0
    # 
    npxs = 3
    upck = '>{0}l'.format(npxs)
    header = _np.array(_struct.unpack(upck, f[ixdr:ixdr + npxs * 4]) )
    ixdr += npxs * 4
    nq, nlb, nm = header
    # 
    npxs = nq
    upck = '>{0}l'.format(npxs)
    header = _np.array(_struct.unpack(upck, f[ixdr:ixdr + npxs * 4]) )
    ixdr += npxs * 4
    # 
    listpar = [[] for i in range(nq)]
    for i in range(nq):
        npxs = header[i]
        upck = '>{0}f'.format(npxs)
        listpar[i] = _np.array(_struct.unpack(upck, f[ixdr:ixdr + npxs * 4]) )
        ixdr += npxs * 4
    # 
    npxs = nlb
    upck = '>{0}f'.format(npxs)
    lbdarr = _np.array(_struct.unpack(upck, f[ixdr:ixdr + npxs * 4]) )
    ixdr += npxs * 4
    # 
    npxs = nm * (nq + nlb)
    upck = '>{0}f'.format(npxs)
    models = _np.array(_struct.unpack(upck, f[ixdr:ixdr + npxs * 4]) )
    ixdr += npxs * 4
    models = models.reshape((nm, -1))
    # this will check if the XDR is finished.
    if ixdr == len(f):
        if not quiet:
            print('# XDR {0} completely read!'.format(xdrpath))
    else:
        _warn.warn('# XDR {0} not completely read!\n# length '
            'difference is {1}'.format(xdrpath, (len(f)-ixdr)/4) )
    # 
    return listpar, lbdarr, models[:, 0:nq], models[:, nq:]


def parnorm(dvals, vmax, vmin_non0, issig0=True, s_non0=0):
    r""" Converts density in normalized range [0-1], and vice-versa.

    If ``issig0``, treats ``r01`` as :math:`\Sigma_0`; otherwise, use it
    as [0-1] value.

    ``s_non0`` forces ``vmin_non0`` to be lower limit of the density scale 
    interval. Example: ``s_non0 = 0.25`` forces the density scale to be 
    between [0.25-1.00].
    """
    dvals = _np.array(dvals)
    if vmin_non0 <= 0 or _np.min(dvals) < 0:
        raise ValueError('`vmin_non0` > 0 and `dvals` >=0 must be True')
    if s_non0 >= 1:
        raise ValueError('`s_non0` <1 must be True')
    if s_non0 > 0:
        vmin_non0 /= (vmax/vmin_non0)**((1/s_non0-1)**-1.)
    if issig0:
        dvals[_np.where(dvals < vmin_non0)] = vmin_non0
        return _np.round( _np.log(dvals/vmin_non0)/_np.log(vmax/vmin_non0), 7 )
    else:
        return _np.round( _np.exp(dvals*_np.log(vmax/vmin_non0))*vmin_non0, 7 )


def densBAnorm(r01, M, issig0=True):
    r""" Converts density in normalized range [0-1], and vice-versa for the 
    BeAtlas.

    If ``issig0``, treats ``r01`` as :math:`\Sigma_0`; otherwise, use it
    as [0-1] value.

    """
    if M < 3.8 or M > 14.6:
        raise ValueError('# Wrong M at bat.normdens() !')
    vmin = 0.02/1.5
    r01 = _np.array(r01)
    if issig0:
        r01[_np.where(r01 < vmin)] = vmin

    # Completo, convergido etapa1, com DEPENDENCIA do vizinho inferior
    # x = [4.2, 4.8, 5.5, 6.4, 7.7, 8.6, 9.6, 10.8]
    # y = [0.05, 0.12, 0.28, 0.28, 0.68, 1.65, 1.65, 4]

    # Completo, convergido etapa2, com DEPENDENCIA do vizinho inferior
    # x = [4.2, 4.8, 5.5, 6.4, 7.7, 8.6, 9.6, 10.8]
    # y = [0.05, 0.12, 0.28, 0.40, 0.68, 1.65, 2.46, 4]

    x = [3.8, 4.2, 4.8, 5.5, 6.4, 7.7, 8.6, 9.6, 10.8, 12.5, 14.6]
    y = [0.05, 0.12, 0.28, 0.40, 0.68, 1.65, 1.65, 4.00, 4.00, 4.00, 4.00]

    d = 7
    x = _np.array(x)  # + 10**-d
    vmax = _np.interp(M, x, y)
    if issig0:
        return _np.round( _np.log(r01/vmin)/_np.log(vmax/vmin), d)
    else:
        return _np.round( _np.exp(r01*_np.log(vmax/vmin))*vmin, d)


def interpolBA2(params, ctrlarr, minfo, models):
    ctrlarr[_np.isnan(ctrlarr)] = params
    return _griddata(minfo, models, ctrlarr)


def interpolBA(params, ctrlarr, lparams, minfo, models, param=True):
    """ Interpola os `modelos` para os parametros `params` 

    | -params = from emcee minimization
    | -ctrlarr = the fixed value of M, ob(W), Z, H, sig, Rd, h, *n*, cos(i).
    |            If it is not fixed, use np.NaN.
    | -Parametric disk model default (`param` == True).

    This function always returns a valid result (i.e., extrapolations from the
    nearest values are always on).

    If it is a 'Non-squared grid' (asymmetric), it will return a zero array if
    a given model is not found.
    """
    nq = 9
    if not param:
        nq = 8
    if len(ctrlarr) != nq:
        raise ValueError('# Wrong ctrlarr format!!')
    params = params[:_np.sum(_np.isnan(ctrlarr))]
    nlb = len(models[0])
    outmodels = _np.empty((2**len(params), nlb))
    mod = BAmod('')
    parlims = _np.zeros((len(params), 2))
    j = 0
    for i in range(nq):
        if ctrlarr[i] is _np.NaN:
            parlims[j] = [_phc.find_nearest(lparams[i], params[j], 
            bigger=False), _phc.find_nearest(lparams[i], params[j], 
            bigger=True)]
            j += 1
    j = 0
    for prod in _product(*parlims):
        allpars = _np.array(ctrlarr)
        idx = _np.isnan(allpars)
        allpars[idx] = prod
        mod.build(allpars, lparams)
        idx = mod.getidx(minfo)
        if _np.sum(idx) == 0:
            return _np.zeros(nlb)
        outmodels[j] = models[idx]
        j += 1
    X0 = parlims[:, 0]
    X1 = parlims[:, 1]
    return _phc.interLinND(params, X0, X1, outmodels)


def check_xdr_limits(xdrlistpar, xdrminfo, xdrmodels, todel=[]):
    """ Check if the XDR file contains models for all parameters within their
    given interval.

    `todel`: list for dimensions to be skipped in the test. 
    `-1` is the observer's inclination angle.
    """
    todel = list(todel)
    notdel = []               
    for i in range(len(xdrlistpar)):
        lvals = _np.unique(xdrminfo[:, i])
        if len(lvals) == 1 or i == len(xdrlistpar)-1:
            print('# Skipping dimension {} of the models!!'.format(i))
            todel.append(i)
        else:
            notdel.append(i)

    minfo = _np.delete(xdrminfo, todel, axis=1)

    # Atualiza lista de parametros e intervalos
    listpar = []
    for i in range(len(minfo[0])):
        arr = _np.unique(minfo[:, i])
        listpar.append(arr)

    lim_chk = [ [_np.min(i), _np.max(i)] for i in listpar]
    i = 0
    for pars in _product(*lim_chk):
        tidx = _np.where( (_np.array(pars) == minfo).all(axis=1) )
        if _np.size(tidx) == 0:
            print('# IMPORTANT MODEL NOT FOUND:')
            print(pars)
            i += 1
    print("# TOTAL OF {} MODELS WEREN'T FOUND!!".format(i))
    return 


# MAIN ###
if __name__ == "__main__":
    pass
